%{
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"

static void yyerror(const char *err);
extern int yylex(void);
extern int lineno(void);
extern void scan_nextfile(char *file);

#define printd(mask, fmt...) if (debug & (mask)) printf(fmt)

#define PRINTD		0x0001
#define DEBUG_PARSE	0x0002
#define DEBUG_OPT	0x0004
#define DEBUG_OPT1	0x0008
#define DEBUG_OPT2	0x0010
#define DEBUG_OPT3	0x0020

static int debug = PRINTD;

struct expr {
	int token;
	union {
		struct expr *expr;
		struct symbol *sym;
	} left, right;
};

static struct expr *if_expr;

struct statement {
	struct statement *next;
	int token;
	char *prompt;
	struct symbol *def;
	struct expr *dep;
	struct expr *dep2;
};

static struct symbol *symbol_hash[256];
static struct symbol symbol_yes = {
	name: "y",
}, symbol_mod = {
	name: "m",
}, symbol_no = {
	name: "n",
};

struct expr *alloc_expr_word(int token, struct symbol *sym)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.sym = sym;
	return e;
}

struct expr *alloc_expr_one(int token, struct expr *ce)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = ce;
	return e;
}

struct expr *alloc_expr_two(int token, struct expr *e1, struct expr *e2)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = e1;
	e->right.expr = e2;
	return e;
}

struct expr *copy_epxr(struct expr *org)
{
	struct expr *e;

	if (!org)
		return NULL;

	e = malloc(sizeof(*org));
	memcpy(e, org, sizeof(*org));
	switch (org->token) {
	case T_WORD:
	case T_WORD_QUOTE:
	case T_WORD_DQUOTE:
		e->left = org->left;
		break;
	case '!':
		e->left.expr = copy_epxr(org->left.expr);
		break;
	case '=':
	case T_UNEQUAL:
	case T_OR:
	case T_AND:
	case '&':
		e->left.expr = copy_epxr(org->left.expr);
		e->right.expr = copy_epxr(org->right.expr);
		break;
	case T_IF:
	case T_ELSE:
		e->left.expr = copy_epxr(org->left.expr);
		e->right.expr = copy_epxr(org->right.expr);
		break;
	default:
		printf("can't copy token %d\n", e->token);
		free(e);
		e = NULL;
		break;
	}

	return e;
}

void free_expr(struct expr *e)
{
	if (!e)
		return;

	switch (e->token) {
	case T_WORD:
	case T_WORD_QUOTE:
	case T_WORD_DQUOTE:
		break;
	case '!':
		free_expr(e->left.expr);
		return;
	case '=':
	case T_UNEQUAL:
	case T_OR:
	case T_AND:
	case '&':
		free_expr(e->left.expr);
		free_expr(e->right.expr);
		break;
	case T_IF:
	case T_ELSE:
		free_expr(e->left.expr);
		free_expr(e->right.expr);
		break;
	default:
		printf("how to free token %d?\n", e->token);
		break;
	}
	free(e);
}

void print_expr(int mask, struct expr *e, int prevtoken)
{
	if (!e) {
		printd(mask, "<none>");
		return;
	}

	switch (e->token) {
	case T_WORD:
	case T_WORD_QUOTE:
	case T_WORD_DQUOTE:
		printd(mask, "%s", e->left.sym->name);
		switch(e->left.sym->type) {
		case SYMBOL_BOOLEAN:
			printd(mask, "?");
			break;
		case SYMBOL_TRISTATE:
			printd(mask, "??");
			break;
		case SYMBOL_DATA:
			printd(mask, "*");
			break;
		}
		break;
	case '!':
		printd(mask, "(!");
		print_expr(mask, e->left.expr, '!');
		printd(mask, ")");
		break;
	case '=':
		print_expr(mask, e->left.expr, '=');
		printd(mask, "=");
		print_expr(mask, e->right.expr, '=');
		break;
	case T_UNEQUAL:
		print_expr(mask, e->left.expr, T_UNEQUAL);
		printd(mask, "!=");
		print_expr(mask, e->right.expr, T_UNEQUAL);
		break;
	case T_OR:
		if (prevtoken != T_OR)
			printd(mask, "(");
		print_expr(mask, e->left.expr, T_OR);
		printd(mask, " || ");
		print_expr(mask, e->right.expr, T_OR);
		if (prevtoken != T_OR)
			printd(mask, ")");
		break;
	case T_AND:
		if (prevtoken != T_AND)
			printd(mask, "(");
		print_expr(mask, e->left.expr, T_AND);
		printd(mask, " && ");
		print_expr(mask, e->right.expr, T_AND);
		if (prevtoken != T_AND)
			printd(mask, ")");
		break;
	case '&':
		//printd(mask, "(");
		print_expr(mask, e->left.expr, '&');
		printd(mask, " & ");
		print_expr(mask, e->right.expr, '&');
		//printd(mask, ")");
		break;
	case '^':
		if (e->left.expr) {
			print_expr(mask, e->left.expr, '^');
			printd(mask, " ^ ");
		}
		printd(mask, "%s", e->right.sym->name);
		break;
	case T_IF:
		if (prevtoken != T_IF)
			printd(mask, "[");
		if (e->right.expr) {
			print_expr(mask, e->right.expr, T_IF);
			printd(mask, " && ");
		}
		print_expr(mask, e->left.expr, T_IF);
		if (prevtoken != T_IF)
			printd(mask, "]");
		break;
	case T_ELSE:
		printd(mask, "[");
		if (e->right.expr) {
			print_expr(mask, e->right.expr, T_ELSE);
			printd(mask, " && ");
		}
		printd(mask, "!");
		print_expr(mask, e->left.expr, T_ELSE);
		printd(mask, "]");
		break;
	default:
		printd(mask, "<unknown token %d>", e->token);
		break;
	}
}

void print_dep(int mask, struct statement *stmt)
{
	printd(mask, "dep: ");
	if (stmt) {
		if (stmt->dep2)
			printd(mask, "[");
		print_expr(mask, stmt->dep, 0);
		if (stmt->dep2) {
			printd(mask, " & ");
			print_expr(mask, stmt->dep2, 0);
			printd(mask, "]");
		}
	} else
		print_expr(mask, if_expr, 0);
	printd(mask, "\n");
}

struct symbol *lookup_symbol(char *name, int type)
{
	struct symbol *symbol;
	char *ptr;
	int hash = 0;

	if (type != SYMBOL_OTHER && !strncmp(name, "CONFIG_", 7)) {
		name += 7;
	} else {
		//printf("Warning: Invalid symbol name '%s'\n", name);
		type = SYMBOL_OTHER;
	}

	for (ptr = name; *ptr; ptr++)
		hash += *ptr;

	for (symbol = symbol_hash[hash&255]; symbol; symbol = symbol->next) {
		if (!strcmp(symbol->name, name))
			return symbol;
	}

	symbol = malloc(sizeof(*symbol));
	memset(symbol, 0, sizeof(*symbol));
	symbol->name = strdup(name);
	symbol->type = type;

	symbol->next = symbol_hash[hash&255];
	symbol_hash[hash&255] = symbol;

	return symbol;
}

struct symbol *lookup_symbol_ref(char *name)
{
	if (name[0] == 'y' && !name[1])
		return &symbol_yes;
	if (name[0] == 'm' && !name[1])
		return &symbol_mod;
	if (name[0] == 'n' && !name[1])
		return &symbol_no;
	if (name[0] == '$')
		return lookup_symbol(name+1, SYMBOL_UNKNOWN);
	return lookup_symbol(name, SYMBOL_OTHER);
}

struct statement *add_stmt(struct symbol *sym, int token, char *prompt, struct symbol *def)
{
	struct statement *stmt = malloc(sizeof(*stmt));
	memset(stmt, 0, sizeof(*stmt));
	stmt->token = token;
	stmt->dep = copy_epxr(if_expr);
	stmt->prompt = prompt;
	stmt->def = def;

	stmt->next = sym->stmt;
	sym->stmt = stmt;
	return stmt;
}

#define YYERROR_VERBOSE
%}

%debug
%verbose
%defines

%union
{
	char *string;
	struct expr *expr;
	struct symbol *symbol;
	struct statement *stmt;
}

%token <string> T_WORD, T_WORD_QUOTE, T_WORD_DQUOTE
%token T_MAINMENU_NAME	"mainmenu_name"
%token T_MAINMENU_OPTION "mainmenu_option"
%token T_NEXT_COMMENT	"next_comment"
%token T_ENDMENU	"endmenu"
%token T_COMMENT	"comment"
%token T_TEXT		"text"
%token T_BOOL		"bool"
%token T_HEX		"hex"
%token T_INT		"int"
%token T_STRING		"string"
%token T_TRISTATE	"tristate"
%token T_DEFINE_BOOL	"define_bool"
%token T_DEFINE_HEX	"define_hex"
%token T_DEFINE_INT	"define_int"
%token T_DEFINE_STRING	"define_string"
%token T_DEFINE_TRISTATE "define_tristate"
%token T_DEP_BOOL	"dep_bool"
%token T_DEP_MBOOL	"dep_mbool"
%token T_DEP_HEX	"dep_hex"
%token T_DEP_INT	"dep_int"
%token T_DEP_STRING	"dep_string"
%token T_DEP_TRISTATE	"dep_tristate"
%token T_UNSET		"unset"
%token T_CHOICE		"choice"
%token T_NCHOICE	"nchoice"
%token T_SOURCE		"source"
%token T_IF		"if"
%token T_ELSE		"else"
%token T_FI		"fi"
%token T_THEN		"then"
%token T_UNEQUAL	"!="
%token T_AND		"-a"
%token T_OR		"-o"

%left T_OR
%left T_AND
%nonassoc '!'
%left '=' T_UNEQUAL

%type <string> prompt, word, menu_comment
%type <symbol> symbol, symbol_ref, symbol_other
%type <stmt> statement
%type <expr> expr, dep

%%
input:	line

line:	/* empty */
{ }

line:	line statement
;

line:	line '\n'
;

statement: "mainmenu_name" prompt '\n'
{
	printd(DEBUG_PARSE, "%d:mainmenu_name: %s\n", lineno(), $2);
}

statement: "comment" prompt '\n'
{
	printd(DEBUG_PARSE, "%d:comment: %s\n", lineno(), $2);
}

statement: "text" prompt '\n'
{
	printd(DEBUG_PARSE, "%d:text: %s\n", lineno(), $2);
}

statement: "bool" prompt symbol '\n'
{
	$$ = add_stmt($3, T_BOOL, $2, NULL);
	printd(DEBUG_PARSE, "%d:bool: %s %s\n", lineno(), $2, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "hex" prompt symbol symbol_other '\n'
{
	$$ = add_stmt($3, T_HEX, $2, $4);
	printd(DEBUG_PARSE, "%d:hex: %s %s %s\n", lineno(), $2, $3->name, $4->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "int" prompt symbol symbol_other '\n'
{
	$$ = add_stmt($3, T_INT, $2, $4);
	printd(DEBUG_PARSE, "%d:int: %s %s %s\n", lineno(), $2, $3->name, $4->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "string" prompt symbol symbol_other '\n'
{
	$$ = add_stmt($3, T_STRING, $2, $4);
	printd(DEBUG_PARSE, "%d:string: %s %s %s\n", lineno(), $2, $3->name, $4->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "tristate" prompt symbol '\n'
{
	$$ = add_stmt($3, T_TRISTATE, $2, NULL);
	printd(DEBUG_PARSE, "%d:tristate: %s %s\n", lineno(), $2, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "define_bool" symbol symbol_ref '\n'
{
	$$ = add_stmt($2, T_DEFINE_BOOL, NULL, $3);
	printd(DEBUG_PARSE, "%d:define_bool: %s %s\n", lineno(), $2->name, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "define_hex" symbol symbol_other '\n'
{
	$$ = add_stmt($2, T_DEFINE_HEX, NULL, $3);
	printd(DEBUG_PARSE, "%d:define_hex: %s %s\n", lineno(), $2->name, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "define_int" symbol symbol_other '\n'
{
	$$ = add_stmt($2, T_DEFINE_INT, NULL, $3);
	printd(DEBUG_PARSE, "%d:define_int: %s %s\n", lineno(), $2->name, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "define_string" symbol symbol_other '\n'
{
	$$ = add_stmt($2, T_DEFINE_STRING, NULL, $3);
	printd(DEBUG_PARSE, "%d:define_string: %s %s\n", lineno(), $2->name, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "define_tristate" symbol symbol_ref '\n'
{
	$$ = add_stmt($2, T_DEFINE_TRISTATE, NULL, $3);
	printd(DEBUG_PARSE, "%d:define_tristate: %s %s\n", lineno(), $2->name, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "dep_bool" prompt symbol dep '\n'
{
	$$ = add_stmt($3, T_DEP_BOOL, $2, NULL);
	$$->dep2 = $4;
	printd(DEBUG_PARSE, "%d:dep_bool: %s %s\n", lineno(), $2, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "dep_mbool" prompt symbol dep '\n'
{
	$$ = add_stmt($3, T_DEP_MBOOL, $2, NULL);
	$$->dep2 = $4;
	printd(DEBUG_PARSE, "%d:dep_mbool: %s %s\n", lineno(), $2, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "dep_hex" prompt symbol symbol_other dep '\n'
{
	$$ = add_stmt($3, T_DEP_HEX, $2, $4);
	$$->dep2 = $5;
	printd(DEBUG_PARSE, "%d:dep_hex: %s %s %s\n", lineno(), $2, $3->name, $4->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "dep_int" prompt symbol symbol_other dep '\n'
{
	$$ = add_stmt($3, T_DEP_INT, $2, $4);
	$$->dep2 = $5;
	printd(DEBUG_PARSE, "%d:dep_int: %s %s %s\n", lineno(), $2, $3->name, $4->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "dep_string" prompt symbol symbol_other dep '\n'
{
	$$ = add_stmt($3, T_DEP_STRING, $2, $4);
	$$->dep2 = $5;
	printd(DEBUG_PARSE, "%d:dep_string: %s %s %s\n", lineno(), $2, $3->name, $4->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "dep_tristate" prompt symbol dep '\n'
{
	$$ = add_stmt($3, T_DEP_TRISTATE, $2, NULL);
	$$->dep2 = $4;
	printd(DEBUG_PARSE, "%d:dep_tristate: %s %s\n", lineno(), $2, $3->name);
	print_dep(DEBUG_PARSE, $$);
}

statement: "unset" symbol '\n'
{
	add_stmt($2, T_UNSET, NULL, NULL);
	printd(DEBUG_PARSE, "%d:unset\n", lineno());
	print_dep(DEBUG_PARSE, $$);
}

statement: "choice" prompt word symbol_other '\n'
{
	struct statement *stmt, *stmt2;
	struct symbol *sym, *def;
	char *ptr, *name, end;

	printd(DEBUG_PARSE, "%d:choice: %s %s %s\n", lineno(), $2, $3, $4->name);
	stmt = malloc(sizeof(*stmt));
	memset(stmt, 0, sizeof(*stmt));
	stmt->token = '^';
	stmt->prompt = $2;
	stmt->def = $4;
	ptr = $3;
	do {
		while (isspace(*ptr))
			ptr++;
		if (!*ptr)
			break;
		name = ptr;
		while (*ptr && !isspace(*ptr))
			ptr++;
		*ptr++ = 0;
		def = lookup_symbol(name, SYMBOL_OTHER);
		while (isspace(*ptr))
			ptr++;
		if (!*ptr)
			break;
		name = ptr;
		while (*ptr && !isspace(*ptr))
			ptr++;
		end = *ptr;
		*ptr++ = 0;
		sym = lookup_symbol(name, SYMBOL_BOOLEAN);
		stmt2 = add_stmt(sym, '^', NULL, def);
		if (!stmt2->next) {
			stmt2->next = stmt;
		} else
			printf("warning: unexpected stmt for xor symbol %s\n", sym->name);
		stmt->dep = alloc_expr_one('^', stmt->dep);
		stmt->dep->right.sym = sym;
		printd(DEBUG_PARSE, "added symbol: %s %s\n", sym->name, stmt2->def->name);
	} while (end);
	print_dep(DEBUG_PARSE, NULL);
}

statement: "nchoice" prompt T_WORD '\n'
{
	printd(DEBUG_PARSE, "%d:nchoice\n", lineno());
	print_dep(DEBUG_PARSE, NULL);
}

statement: "if" '[' expr ']' semi_or_nl "then" '\n'
{
	struct expr *e;
	printd(DEBUG_PARSE, "%d:if ", lineno());
	print_expr(DEBUG_PARSE, $3, 0);
	printd(DEBUG_PARSE, "\n");
	e = alloc_expr_one(T_IF, $3);
	e->right.expr = if_expr;
	if_expr = e;
	$$ = NULL;
}

semi_or_nl: ';' | '\n'

statement: "else" '\n'
{
	printd(DEBUG_PARSE, "%d:else\n", lineno());
	if_expr->token = T_ELSE;
	$$ = NULL;
}

statement: "fi" '\n'
{
	struct expr *e;
	printd(DEBUG_PARSE, "%d:fi\n", lineno());
	e = if_expr;
	if (e) {
		if_expr = e->right.expr;
		free_expr(e->left.expr);
		free(e);
	} else
		YYABORT;
	$$ = NULL;
}

statement: "mainmenu_option" "next_comment" menu_comment  '\n'
{
	printd(DEBUG_PARSE, "%d:mainmenu_option: %s\n", lineno(), $3);
}

menu_comment: '\n' "comment" prompt { $$ = $3; }
	| '\n' menu_comment { $$ = $2; }

statement: "endmenu" '\n'
{
	printd(DEBUG_PARSE, "%d:endmenu\n", lineno());
}

statement: "source" word '\n'
{
	printd(DEBUG_PARSE, "%d:source: %s\n", lineno(), $2);
	scan_nextfile($2);
}

prompt:   T_WORD_QUOTE
	| T_WORD_DQUOTE

word:	  T_WORD
	| T_WORD_QUOTE
	| T_WORD_DQUOTE

dep:	  symbol_ref		{ $$ = alloc_expr_word(T_WORD, $1); }
	| dep symbol_ref	{ $$ = alloc_expr_two('&', $1, alloc_expr_word(T_WORD, $2)); }

symbol:	T_WORD			{ $$ = lookup_symbol($1, SYMBOL_UNKNOWN); free($1); }

expr:	  symbol_ref		{ $$ = alloc_expr_word(T_WORD, $1); }
	| '!' expr		{ $$ = alloc_expr_one('!', $2); }
	| expr '=' expr		{ $$ = alloc_expr_two('=', $1, $3); }
	| expr "!=" expr	{ $$ = alloc_expr_two(T_UNEQUAL, $1, $3); }
	| expr "-a" expr	{ $$ = alloc_expr_two(T_AND, $1, $3); }
	| expr "-o" expr	{ $$ = alloc_expr_two(T_OR, $1, $3); }

symbol_ref: T_WORD		{ $$ = lookup_symbol_ref($1); free($1); }
	  | T_WORD_QUOTE	{ $$ = lookup_symbol_ref($1); free($1); }
	  | T_WORD_DQUOTE	{ $$ = lookup_symbol_ref($1); free($1); }

symbol_other: T_WORD		{ $$ = lookup_symbol($1, SYMBOL_OTHER); free($1); }
	    | T_WORD_QUOTE	{ $$ = lookup_symbol($1, SYMBOL_OTHER); free($1); }
	    | T_WORD_DQUOTE	{ $$ = lookup_symbol($1, SYMBOL_OTHER); free($1); }

%%
static void yyerror(const char *err)
{
	printf("%d:%s\n", lineno(), err);
	exit(1);
}

#define for_all_symbols for (i = 0; i < 256; i++) for (sym = symbol_hash[i]; sym; sym = sym->next) if (sym->type != SYMBOL_OTHER)

void print_symbols(void)
{
	struct symbol *sym;
	struct statement *stmt;
	int i;

	for_all_symbols {
		printf("config: %s\n", sym->name);
		for (stmt = sym->stmt; stmt; stmt = stmt->next) {
			switch (stmt->token) {
			case T_BOOL: printf("  bool\n"); break;
			case T_HEX: printf("  hex\n"); break;
			case T_INT: printf("  int\n"); break;
			case T_STRING: printf("  string\n"); break;
			case T_TRISTATE: printf("  tristate\n"); break;
			case T_DEFINE_BOOL: printf("  define_bool\n"); break;
			case T_DEFINE_HEX: printf("  define_hex\n"); break;
			case T_DEFINE_INT: printf("  define_int\n"); break;
			case T_DEFINE_STRING: printf("  define_string\n"); break;
			case T_DEFINE_TRISTATE: printf("  define_tristate\n"); break;
			case T_DEP_BOOL: printf("  dep_bool\n"); break;
			case T_DEP_MBOOL: printf("  dep_mbool\n"); break;
			case T_DEP_HEX: printf("  dep_hex\n"); break;
			case T_DEP_INT: printf("  dep_int\n"); break;
			case T_DEP_STRING: printf("  dep_string\n"); break;
			case T_DEP_TRISTATE: printf("  dep_tristate\n"); break;
			case T_UNSET: printf("  unset\n"); break;
			case '^': printf("  choice\n"); break;
			default: printf("  unknown\n");
			}
			if (stmt->prompt)
				printf("    prompt: %s\n", stmt->prompt);
			if (stmt->def)
				printf("    default: %s\n", stmt->def->name);
			printf("    dep: ");
			if (stmt->dep2)
				printf("[");
			print_expr(PRINTD, stmt->dep, 0);
			if (stmt->dep2) {
				printf(" & ");
				print_expr(PRINTD, stmt->dep2, 0);
				printf("]");
			}
			printf("\n");
		}
		if (sym->help) {
			printf("  help:\n%s", sym->help);
		}
		printf("\n");
	}
}

#define SKIP_OPT(e)		\
	if (!e)			\
		return NULL;	\
	switch (e->token) {	\
	case T_WORD:		\
	case T_WORD_QUOTE:	\
	case T_WORD_DQUOTE:	\
	case '^':		\
		return e;	\
	}

struct expr *optimize_expr1(struct expr *e)
{
	struct expr *tmp;

	SKIP_OPT(e);
	switch (e->token) {
	case T_IF:
		print_expr(DEBUG_OPT1, e, 0);
		printd(DEBUG_OPT1, " -> ");
		tmp = e->left.expr;
		if (!e->right.expr) {
			free(e);
			e = optimize_expr1(tmp);
			print_expr(DEBUG_OPT1, e, 0);
			printd(DEBUG_OPT1, "\n");
			return e;
		}
		e->left.expr = e->right.expr;
		e->right.expr = tmp;
		e->token = T_AND;
		print_expr(DEBUG_OPT1, e, 0);
		printd(DEBUG_OPT1, "\n");
		break;
	case T_ELSE:
		print_expr(DEBUG_OPT1, e, 0);
		printd(DEBUG_OPT1, " -> ");
		e->token = '!';
		if (!e->right.expr) {
			print_expr(DEBUG_OPT1, e, 0);
			printd(DEBUG_OPT1, "\n");
			break;
		}
		tmp = e;
		e = alloc_expr_two(T_AND, e->right.expr, e);
		tmp->right.expr = NULL;
		print_expr(DEBUG_OPT1, e, 0);
		printd(DEBUG_OPT1, "\n");
		break;
	case '&':
		print_expr(DEBUG_OPT1, e, 0);
		printd(DEBUG_OPT1, " -> ");
		e->token = T_AND;
		print_expr(DEBUG_OPT1, e, 0);
		printd(DEBUG_OPT1, "\n");
		break;
	case '=':
	case T_UNEQUAL:
		if (e->left.expr->token != T_WORD)
			break;
		if (e->right.expr->token != T_WORD)
			break;
		tmp = e->left.expr;
		if (tmp->left.sym == &symbol_yes ||
		    tmp->left.sym == &symbol_mod ||
		    tmp->left.sym == &symbol_no) {
			print_expr(DEBUG_OPT1, e, 0);
			printd(DEBUG_OPT1, " -> ");
			e->left.expr = e->right.expr;
			e->right.expr = tmp;
			print_expr(DEBUG_OPT1, e, 0);
			printd(DEBUG_OPT1, "\n");
		}
		break;
	}
	e->left.expr = optimize_expr1(e->left.expr);
	e->right.expr = optimize_expr1(e->right.expr);
	return e;
}

struct expr *optimize_expr2_bool(struct expr *e)
{
	struct expr *tmp;
	struct symbol *sym;

	SKIP_OPT(e);

	e->left.expr = optimize_expr2_bool(e->left.expr);
	e->right.expr = optimize_expr2_bool(e->right.expr);

	switch (e->token) {
	case '=':
		if (e->left.expr->token != T_WORD)
			break;
		if (e->right.expr->token != T_WORD)
			break;
		tmp = e->left.expr;
		if (tmp->left.sym->type != SYMBOL_BOOLEAN)
			break;
		sym = e->right.expr->left.sym;
		if (sym == &symbol_yes) {
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, " -> ");
			free(e);
			e = tmp;
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, "\n");
			return e;
		}
		if (sym == &symbol_no) {
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, " -> ");
			e->token = '!';
			free(e->right.expr);
			e->right.expr = NULL;
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, "\n");
			return e;
		}
		break;
	case T_UNEQUAL:
		if (e->left.expr->token != T_WORD)
			break;
		if (e->right.expr->token != T_WORD)
			break;
		tmp = e->left.expr;
		if (tmp->left.sym->type != SYMBOL_BOOLEAN)
			break;
		sym = e->right.expr->left.sym;
		if (sym == &symbol_no) {
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, " -> ");
			free(e);
			e = tmp;
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, "\n");
			return e;
		}
		if (sym == &symbol_yes) {
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, " -> ");
			e->token = '!';
			free(e->right.expr);
			e->right.expr = NULL;
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, "\n");
			return e;
		}
		break;
	case '!':
		switch (e->left.expr->token) {
		case '!':
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, " -> ");
			tmp = e->left.expr->left.expr;
			free(e->left.expr);
			free(e);
			e = tmp;
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, "\n");
			break;
		case '=':
		case T_UNEQUAL:
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, " -> ");
			tmp = e->left.expr;
			free(e);
			e = tmp;
			if (e->token == '=')
				e->token = T_UNEQUAL;
			else
				e->token = '=';
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, "\n");
			break;
		case T_OR:
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, " -> ");
			tmp = e->left.expr;
			e->token = T_AND;
			e->right.expr = alloc_expr_one('!', tmp->right.expr);
			tmp->token = '!';
			tmp->right.expr = NULL;
			print_expr(DEBUG_OPT2, e, 0);
			printd(DEBUG_OPT2, "\n");
			e = optimize_expr2_bool(e);
			break;
		}
		break;
	}
	return e;
}

int dep_check_symbol(struct symbol *sym);

int dep_check_expr(struct expr *e)
{
	if (!e)
		return 0;
	switch (e->token) {
	case T_WORD:
	case T_WORD_QUOTE:
	case T_WORD_DQUOTE:
		return dep_check_symbol(e->left.sym);
	}
	return dep_check_expr(e->left.expr) || dep_check_expr(e->right.expr);
}

int dep_check_symbol(struct symbol *sym)
{
	struct statement *stmt;
	int res = 0;

	if (sym->flags & SYMBOL_CHECK) {
		printf("recursive dependency: %s", sym->name);
		return 1;
	}

	sym->flags |= SYMBOL_CHECK;
	for (stmt = sym->stmt; stmt; stmt = stmt->next) {
		if (dep_check_expr(stmt->dep)) {
			printf(" %s", sym->name);
			res = 1;
			break;
		}

		if (stmt->token == '^')
			break;
	}
	sym->flags &= ~SYMBOL_CHECK;

	return res;
}

void optimize_config(void)
{
	struct symbol *sym;
	struct statement *stmt;
	int i;

	for_all_symbols {
		printd(DEBUG_OPT1, "o1: %s\n", sym->name);
		for (stmt = sym->stmt; stmt; stmt = stmt->next) {
			switch (stmt->token) {
			case '^':
			case T_BOOL:
			case T_DEP_BOOL:
			case T_DEP_MBOOL:
				if (sym->type == SYMBOL_UNKNOWN)
					sym->type = SYMBOL_BOOLEAN;
				if (sym->type != SYMBOL_BOOLEAN)
					printf("warning: symbol %s has conflicting types %d,%d\n", sym->name, sym->type, SYMBOL_BOOLEAN);
				break;
			case T_HEX:
			case T_INT:
			case T_STRING:
			case T_DEP_HEX:
			case T_DEP_INT:
			case T_DEP_STRING:
				if (sym->type == SYMBOL_UNKNOWN)
					sym->type = SYMBOL_DATA;
				if (sym->type != SYMBOL_DATA)
					printf("warning: symbol %s has conflicting types %d,%d\n", sym->name, sym->type, SYMBOL_DATA);
				break;
			case T_TRISTATE:
			case T_DEP_TRISTATE:
				if (sym->type == SYMBOL_UNKNOWN)
					sym->type = SYMBOL_TRISTATE;
				if (sym->type != SYMBOL_TRISTATE)
					printf("warning: symbol %s has conflicting types %d,%d\n", sym->name, sym->type, SYMBOL_TRISTATE);
				break;
			}
			print_expr(DEBUG_OPT1, stmt->dep, 0);
			printd(DEBUG_OPT1, " & ");
			print_expr(DEBUG_OPT1, stmt->dep2, 0);
			printd(DEBUG_OPT1, "\n");
			stmt->dep = optimize_expr1(stmt->dep);
			if (stmt->dep || stmt->dep2) {
				if (stmt->dep && stmt->dep2) {
					stmt->dep = alloc_expr_two(T_AND, stmt->dep, optimize_expr1(stmt->dep2));
				} else if (stmt->dep2)
					stmt->dep = optimize_expr1(stmt->dep2);
				stmt->dep2 = NULL;
			}
			print_expr(DEBUG_OPT1, stmt->dep, 0);
			printd(DEBUG_OPT1, "\n");
		}
		for (stmt = sym->stmt; stmt; stmt = stmt->next) {
			switch (stmt->token) {
			case T_DEFINE_BOOL:
				if (sym->type == SYMBOL_UNKNOWN)
					sym->type = SYMBOL_BOOLEAN;
				if (sym->type != SYMBOL_BOOLEAN)
					printf("warning: symbol %s has conflicting types %d,%d\n", sym->name, sym->type, SYMBOL_BOOLEAN);
				break;
			case T_DEFINE_TRISTATE:
				if (sym->type == SYMBOL_UNKNOWN)
					sym->type = SYMBOL_TRISTATE;
				if (sym->type != SYMBOL_TRISTATE)
					printf("warning: symbol %s has conflicting types %d,%d\n", sym->name, sym->type, SYMBOL_TRISTATE);
				break;
			case T_DEFINE_HEX:
			case T_DEFINE_INT:
			case T_DEFINE_STRING:
				if (sym->type == SYMBOL_UNKNOWN)
					sym->type = SYMBOL_DATA;
				if (sym->type != SYMBOL_DATA)
					printf("warning: symbol %s has conflicting types %d,%d\n", sym->name, sym->type, SYMBOL_DATA);
				break;
			}
		}
	}
	for_all_symbols {
		printd(DEBUG_OPT2, "o2: %s\n", sym->name);
		for (stmt = sym->stmt; stmt; stmt = stmt->next) {
			print_expr(DEBUG_OPT2, stmt->dep, 0);
			printd(DEBUG_OPT2, "\n");
			stmt->dep = optimize_expr2_bool(stmt->dep);
			print_expr(DEBUG_OPT2, stmt->dep, 0);
			printd(DEBUG_OPT2, "\n");
		}
	}
	for_all_symbols {
		if (dep_check_symbol(sym))
			printf("\n");
	}
}
